@@ -1,63 +1,22 @@
-package org.apache.regexp;
-
 /*
- * ====================================================================
+ * Licensed to the Apache Software Foundation (ASF) under one or more
+ * contributor license agreements.  See the NOTICE file distributed with
+ * this work for additional information regarding copyright ownership.
+ * The ASF licenses this file to You under the Apache License, Version 2.0
+ * (the "License"); you may not use this file except in compliance with
+ * the License.  You may obtain a copy of the License at
  *
- * The Apache Software License, Version 1.1
- *
- * Copyright (c) 1999-2003 The Apache Software Foundation.  All rights
- * reserved.
- *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
- *
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
- *
- * 2. Redistributions in binary form must reproduce the above copyright
- *    notice, this list of conditions and the following disclaimer in
- *    the documentation and/or other materials provided with the
- *    distribution.
- *
- * 3. The end-user documentation included with the redistribution, if
- *    any, must include the following acknowlegement:
- *       "This product includes software developed by the
- *        Apache Software Foundation (http://www.apache.org/)."
- *    Alternately, this acknowlegement may appear in the software itself,
- *    if and wherever such third-party acknowlegements normally appear.
- *
- * 4. The names "The Jakarta Project", "Jakarta-Regexp", and "Apache Software
- *    Foundation" must not be used to endorse or promote products derived
- *    from this software without prior written permission. For written
- *    permission, please contact apache@apache.org.
- *
- * 5. Products derived from this software may not be called "Apache"
- *    nor may "Apache" appear in their names without prior written
- *    permission of the Apache Group.
- *
- * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
- * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
- * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
- * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
- * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
- * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
- * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
- * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
- * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
- * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
- * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
- * ====================================================================
- *
- * This software consists of voluntary contributions made by many
- * individuals on behalf of the Apache Software Foundation.  For more
- * information on the Apache Software Foundation, please see
- * <http://www.apache.org/>.
+ *     http://www.apache.org/licenses/LICENSE-2.0
  *
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
  */
 
-import org.apache.regexp.RE;
+package org.apache.regexp;
+
 import java.util.Hashtable;
 
 /**
@@ -72,7 +31,7 @@ import java.util.Hashtable;
  *
  * @author <a href="mailto:jonl@muppetlabs.com">Jonathan Locke</a>
  * @author <a href="mailto:gholam@xtra.co.nz">Michael McCallum</a>
- * @version $Id: RECompiler.java,v 1.11 2003/09/01 18:31:53 vgritsenko Exp $
+ * @version $Id: RECompiler.java 518156 2007-03-14 14:31:26Z vgritsenko $
  */
 public class RECompiler
 {
@@ -92,22 +51,18 @@ public class RECompiler
     static final int NODE_TOPLEVEL = 2;                 // True if top level expr
 
     // Special types of 'escapes'
-    static final char ESC_MASK     = 0xfff0;            // Escape complexity mask
-    static final char ESC_BACKREF  = 0xffff;            // Escape is really a backreference
-    static final char ESC_COMPLEX  = 0xfffe;            // Escape isn't really a true character
-    static final char ESC_CLASS    = 0xfffd;            // Escape represents a whole class of characters
+    static final int ESC_MASK      = 0xffff0;           // Escape complexity mask
+    static final int ESC_BACKREF   = 0xfffff;           // Escape is really a backreference
+    static final int ESC_COMPLEX   = 0xffffe;           // Escape isn't really a true character
+    static final int ESC_CLASS     = 0xffffd;           // Escape represents a whole class of characters
 
     // {m,n} stacks
-    int maxBrackets = 10;                               // Maximum number of bracket pairs
     static final int bracketUnbounded = -1;             // Unbounded value
-    int brackets = 0;                                   // Number of bracket sets
-    int[] bracketStart = null;                          // Starting point
-    int[] bracketEnd = null;                            // Ending point
-    int[] bracketMin = null;                            // Minimum number of matches
-    int[] bracketOpt = null;                            // Additional optional matches
+    int bracketMin;                                     // Minimum number of matches
+    int bracketOpt;                                     // Additional optional matches
 
     // Lookup table for POSIX character class names
-    static Hashtable hashPOSIX = new Hashtable();
+    static final Hashtable hashPOSIX = new Hashtable();
     static
     {
         hashPOSIX.put("alnum",     new Character(RE.POSIX_CLASS_ALNUM));
@@ -189,9 +144,9 @@ public class RECompiler
 
         // Move everything from insertAt to the end down nodeSize elements
         System.arraycopy(instruction, insertAt, instruction, insertAt + RE.nodeSize, lenInstruction - insertAt);
-        instruction[insertAt + RE.offsetOpcode] = opcode;
-        instruction[insertAt + RE.offsetOpdata] = (char)opdata;
-        instruction[insertAt + RE.offsetNext] = 0;
+        instruction[insertAt /* + RE.offsetOpcode */] = opcode;
+        instruction[insertAt    + RE.offsetOpdata   ] = (char) opdata;
+        instruction[insertAt    + RE.offsetNext     ] = 0;
         lenInstruction += RE.nodeSize;
     }
 
@@ -211,20 +166,29 @@ public class RECompiler
             // if the node we are supposed to point to is in the chain then
             // point to the end of the program instead.
             // Michael McCallum <gholam@xtra.co.nz>
-            // FIXME: // This is a _hack_ to stop infinite programs.
+            // FIXME: This is a _hack_ to stop infinite programs.
             // I believe that the implementation of the reluctant matches is wrong but
             // have not worked out a better way yet.
-            if ( node == pointTo ) {
-              pointTo = lenInstruction;
+            if (node == pointTo) {
+                pointTo = lenInstruction;
             }
             node += next;
             next = instruction[node + RE.offsetNext];
         }
+
         // if we have reached the end of the program then dont set the pointTo.
         // im not sure if this will break any thing but passes all the tests.
         if ( node < lenInstruction ) {
+            // Some patterns result in very large programs which exceed
+            // capacity of the short used for specifying signed offset of the
+            // next instruction. Example: a{1638}
+            int offset = pointTo - node;
+            if (offset != (short) offset) {
+                throw new RESyntaxException("Exceeded short jump range.");
+            }
+
             // Point the last node in the chain to pointTo.
-            instruction[node + RE.offsetNext] = (char)(short)(pointTo - node);
+            instruction[node + RE.offsetNext] = (char) (short) offset;
         }
     }
 
@@ -240,9 +204,9 @@ public class RECompiler
         ensure(RE.nodeSize);
 
         // Add new node at end
-        instruction[lenInstruction + RE.offsetOpcode] = opcode;
-        instruction[lenInstruction + RE.offsetOpdata] = (char)opdata;
-        instruction[lenInstruction + RE.offsetNext] = 0;
+        instruction[lenInstruction /* + RE.offsetOpcode */] = opcode;
+        instruction[lenInstruction    + RE.offsetOpdata   ] = (char) opdata;
+        instruction[lenInstruction    + RE.offsetNext     ] = 0;
         lenInstruction += RE.nodeSize;
 
         // Return index of new node
@@ -268,55 +232,6 @@ public class RECompiler
         throw new RESyntaxException(s);
     }
 
-    /**
-     * Allocate storage for brackets only as needed
-     */
-    void allocBrackets()
-    {
-        // Allocate bracket stacks if not already done
-        if (bracketStart == null)
-        {
-            // Allocate storage
-            bracketStart = new int[maxBrackets];
-            bracketEnd   = new int[maxBrackets];
-            bracketMin   = new int[maxBrackets];
-            bracketOpt   = new int[maxBrackets];
-
-            // Initialize to invalid values
-            for (int i = 0; i < maxBrackets; i++)
-            {
-                bracketStart[i] = bracketEnd[i] = bracketMin[i] = bracketOpt[i] = -1;
-            }
-        }
-    }
-
-    /** Enlarge storage for brackets only as needed. */
-    synchronized void reallocBrackets() {
-        // trick the tricky
-        if (bracketStart == null) {
-            allocBrackets();
-        }
-
-        int new_size = maxBrackets * 2;
-        int[] new_bS = new int[new_size];
-        int[] new_bE = new int[new_size];
-        int[] new_bM = new int[new_size];
-        int[] new_bO = new int[new_size];
-        // Initialize to invalid values
-        for (int i=brackets; i<new_size; i++) {
-            new_bS[i] = new_bE[i] = new_bM[i] = new_bO[i] = -1;
-        }
-        System.arraycopy(bracketStart,0, new_bS,0, brackets);
-        System.arraycopy(bracketEnd,0,   new_bE,0, brackets);
-        System.arraycopy(bracketMin,0,   new_bM,0, brackets);
-        System.arraycopy(bracketOpt,0,   new_bO,0, brackets);
-        bracketStart = new_bS;
-        bracketEnd   = new_bE;
-        bracketMin   = new_bM;
-        bracketOpt   = new_bO;
-        maxBrackets  = new_size;
-    }
-
     /**
      * Match bracket {m,n} expression put results in bracket member variables
      * @exception RESyntaxException Thrown if the regular expression has invalid syntax.
@@ -343,7 +258,7 @@ public class RECompiler
         }
         try
         {
-            bracketMin[brackets] = Integer.parseInt(number.toString());
+            bracketMin = Integer.parseInt(number.toString());
         }
         catch (NumberFormatException e)
         {
@@ -360,7 +275,7 @@ public class RECompiler
         if (pattern.charAt(idx) == '}')
         {
             idx++;
-            bracketOpt[brackets] = 0;
+            bracketOpt = 0;
             return;
         }
 
@@ -380,7 +295,7 @@ public class RECompiler
         if (pattern.charAt(idx) == '}')
         {
             idx++;
-            bracketOpt[brackets] = bracketUnbounded;
+            bracketOpt = bracketUnbounded;
             return;
         }
 
@@ -398,7 +313,7 @@ public class RECompiler
         }
         try
         {
-            bracketOpt[brackets] = Integer.parseInt(number.toString()) - bracketMin[brackets];
+            bracketOpt = Integer.parseInt(number.toString()) - bracketMin;
         }
         catch (NumberFormatException e)
         {
@@ -406,7 +321,7 @@ public class RECompiler
         }
 
         // Optional repetitions must be >= 0
-        if (bracketOpt[brackets] < 0)
+        if (bracketOpt < 0)
         {
             syntaxError("Bad range");
         }
@@ -427,7 +342,7 @@ public class RECompiler
      * @return ESC_* code or character if simple escape
      * @exception RESyntaxException Thrown if the regular expression has invalid syntax.
      */
-    char escape() throws RESyntaxException
+    int escape() throws RESyntaxException
     {
         // "Shouldn't" happen
         if (pattern.charAt(idx) != '\\')
@@ -494,7 +409,7 @@ public class RECompiler
                             }
                         }
                     }
-                    return (char)val;
+                    return val;
                 }
 
             case 't':
@@ -533,7 +448,7 @@ public class RECompiler
                             val = ((val << 3) + (pattern.charAt(idx++) - '0'));
                         }
                     }
-                    return (char)val;
+                    return val;
                 }
 
                 // It's actually a backreference (\[1-9]), not an escape
@@ -605,7 +520,7 @@ public class RECompiler
         // Parse class declaration
         char CHAR_INVALID = Character.MAX_VALUE;
         char last = CHAR_INVALID;
-        char simpleChar = 0;
+        char simpleChar;
         boolean include = true;
         boolean definingRange = false;
         int idxFirst = idx;
@@ -632,7 +547,7 @@ public class RECompiler
                 case '\\':
                 {
                     // Escape always advances the stream
-                    char c;
+                    int c;
                     switch (c = escape ())
                     {
                         case ESC_COMPLEX:
@@ -653,9 +568,24 @@ public class RECompiler
                             switch (pattern.charAt(idx - 1))
                             {
                                 case RE.E_NSPACE:
-                                case RE.E_NDIGIT:
+                                    range.include(Character.MIN_VALUE, 7, include);   // [Min - \b )
+                                    range.include((char) 11, include);                // ( \n - \f )
+                                    range.include(14, 31, include);                   // ( \r - ' ')
+                                    range.include(33, Character.MAX_VALUE, include);  // (' ' - Max]
+                                    break;
+
                                 case RE.E_NALNUM:
-                                    syntaxError("Bad character class");
+                                    range.include(Character.MIN_VALUE, '/', include); // [Min - '0')
+                                    range.include(':', '@', include);                 // ('9' - 'A')
+                                    range.include('[', '^', include);                 // ('Z' - '_')
+                                    range.include('`', include);                      // ('_' - 'a')
+                                    range.include('{', Character.MAX_VALUE, include); // ('z' - Max]
+                                    break;
+
+                                case RE.E_NDIGIT:
+                                    range.include(Character.MIN_VALUE, '/', include); // [Min - '0')
+                                    range.include(':', Character.MAX_VALUE, include); // ('9' - Max]
+                                    break;
 
                                 case RE.E_SPACE:
                                     range.include('\t', include);
@@ -685,7 +615,7 @@ public class RECompiler
                         default:
 
                             // Escape is simple so treat as a simple char
-                            simpleChar = c;
+                            simpleChar = (char) c;
                             break switchOnCharacter;
                     }
                 }
@@ -850,7 +780,7 @@ public class RECompiler
                     {
                         // Get the escaped character (advances input automatically)
                         int idxBeforeEscape = idx;
-                        char c = escape();
+                        int c = escape();
 
                         // Check if it's a simple escape (as opposed to, say, a backreference)
                         if ((c & ESC_MASK) == ESC_MASK)
@@ -861,7 +791,7 @@ public class RECompiler
                         }
 
                         // Add escaped char to atom
-                        emit(c);
+                        emit((char) c);
                         lenAtom++;
                     }
                     break;
@@ -991,6 +921,7 @@ public class RECompiler
         {
             return ret;
         }
+
         boolean greedy = true;
         char closureType = pattern.charAt(idx);
         switch (closureType)
@@ -1001,15 +932,19 @@ public class RECompiler
                 // The current node can be null
                 flags[0] |= NODE_NULLABLE;
 
+                // Drop through
+
             case '+':
 
                 // Eat closure character
                 idx++;
 
+                // Drop through
+
             case '{':
 
                 // Don't allow blantant stupidity
-                int opcode = instruction[ret + RE.offsetOpcode];
+                int opcode = instruction[ret /* + RE.offsetOpcode */];
                 if (opcode == RE.OP_BOL || opcode == RE.OP_EOL)
                 {
                     syntaxError("Bad closure operand");
@@ -1018,7 +953,6 @@ public class RECompiler
                 {
                     syntaxError("Closure operand can't be nullable");
                 }
-                break;
         }
 
         // If the next character is a '?', make the closure non-greedy (reluctant)
@@ -1035,163 +969,141 @@ public class RECompiler
             {
                 case '{':
                 {
-                    // We look for our bracket in the list
-                    boolean found = false;
-                    int i;
-                    allocBrackets();
-                    for (i = 0; i < brackets; i++)
-                    {
-                        if (bracketStart[i] == idx)
-                        {
-                            found = true;
-                            break;
-                        }
-                    }
+                    bracket();
+                    int bracketEnd = idx;
+                    int bracketMin = this.bracketMin;
+                    int bracketOpt = this.bracketOpt;
 
-                    // If its not in the list we parse the {m,n}
-                    if (!found)
-                    {
-                        if (brackets >= maxBrackets)
-                        {
-                            reallocBrackets();
-                        }
-                        bracketStart[brackets] = idx;
-                        bracket();
-                        bracketEnd[brackets] = idx;
-                        i = brackets++;
-                    }
+                    // Pointer to the last terminal
+                    int pos = ret;
 
                     // Process min first
-                    if (bracketMin[i]-- > 0)
+                    for (int c = 0; c < bracketMin; c++)
                     {
-                        if (bracketMin[i] > 0 || bracketOpt[i] != 0) {
-                            // Rewind stream and run it through again - more matchers coming
-                            idx = idxBeforeTerminal;
-                        } else {
-                            // Bug #1030: No optinal matches - no need to rewind
-                            idx = bracketEnd[i];
-                        }
-                        break;
+                        // Rewind stream and run it through again - more matchers coming
+                        idx = idxBeforeTerminal;
+                        setNextOfEnd(pos, pos = terminal(terminalFlags));
                     }
 
                     // Do the right thing for maximum ({m,})
-                    if (bracketOpt[i] == bracketUnbounded)
+                    if (bracketOpt == bracketUnbounded)
                     {
                         // Drop through now and closure expression.
                         // We are done with the {m,} expr, so skip rest
-                        closureType = '*';
-                        bracketOpt[i] = 0;
-                        idx = bracketEnd[i];
-                    }
-                    else
-                        if (bracketOpt[i]-- > 0)
-                        {
-                            if (bracketOpt[i] > 0)
-                            {
-                                // More optional matchers - 'play it again sam!'
-                                idx = idxBeforeTerminal;
-                            } else {
-                                // Bug #1030: We are done - this one is last and optional
-                                idx = bracketEnd[i];
-                            }
-                            // Drop through to optionally close
-                            closureType = '?';
-                        }
-                        else
-                        {
-                            // Rollback terminal - neither min nor opt matchers present
-                            lenInstruction = ret;
-                            node(RE.OP_NOTHING, 0);
-
-                            // We are done. skip the rest of {m,n} expr
-                            idx = bracketEnd[i];
-                            break;
-                        }
-                }
-
-                // Fall through!
-
-                case '?':
-                case '*':
-
-                    if (!greedy)
-                    {
+                        idx = bracketEnd;
+                        nodeInsert(RE.OP_STAR, 0, pos);
+                        setNextOfEnd(pos + RE.nodeSize, pos);
                         break;
                     }
-
-                    if (closureType == '?')
+                    else if (bracketOpt > 0)
                     {
-                        // X? is compiled as (X|)
-                        nodeInsert(RE.OP_BRANCH, 0, ret);                 // branch before X
-                        setNextOfEnd(ret, node (RE.OP_BRANCH, 0));        // inserted branch to option
-                        int nothing = node (RE.OP_NOTHING, 0);            // which is OP_NOTHING
-                        setNextOfEnd(ret, nothing);                       // point (second) branch to OP_NOTHING
-                        setNextOfEnd(ret + RE.nodeSize, nothing);         // point the end of X to OP_NOTHING node
+                        int opt[] = new int[bracketOpt + 1];
+                        // Surround first optional terminal with MAYBE
+                        nodeInsert(RE.OP_MAYBE, 0, pos);
+                        opt[0] = pos;
+
+                        // Add all the rest optional terminals with preceeding MAYBEs
+                        for (int c = 1; c < bracketOpt; c++)
+                        {
+                            opt[c] = node(RE.OP_MAYBE, 0);
+                            // Rewind stream and run it through again - more matchers coming
+                            idx = idxBeforeTerminal;
+                            terminal(terminalFlags);
+                        }
+
+                        // Tie ends together
+                        int end = opt[bracketOpt] = node(RE.OP_NOTHING, 0);
+                        for (int c = 0; c < bracketOpt; c++)
+                        {
+                            setNextOfEnd(opt[c], end);
+                            setNextOfEnd(opt[c] + RE.nodeSize, opt[c + 1]);
+                        }
+                    }
+                    else
+                    {
+                        // Rollback terminal - no opt matchers present
+                        lenInstruction = pos;
+                        node(RE.OP_NOTHING, 0);
                     }
 
-                    if (closureType == '*')
-                    {
-                        // X* is compiled as (X{gotoX}|)
-                        nodeInsert(RE.OP_BRANCH, 0, ret);                         // branch before X
-                        setNextOfEnd(ret + RE.nodeSize, node(RE.OP_BRANCH, 0));   // end of X points to an option
-                        setNextOfEnd(ret + RE.nodeSize, node(RE.OP_GOTO, 0));     // to goto
-                        setNextOfEnd(ret + RE.nodeSize, ret);                     // the start again
-                        setNextOfEnd(ret, node(RE.OP_BRANCH, 0));                 // the other option is
-                        setNextOfEnd(ret, node(RE.OP_NOTHING, 0));                // OP_NOTHING
-                    }
+                    // We are done. skip the reminder of {m,n} expr
+                    idx = bracketEnd;
                     break;
+                }
+
+                case '?':
+                {
+                    nodeInsert(RE.OP_MAYBE, 0, ret);
+                    int n = node(RE.OP_NOTHING, 0);
+                    setNextOfEnd(ret, n);
+                    setNextOfEnd(ret + RE.nodeSize, n);
+                    break;
+                }
+
+                case '*':
+                {
+                    nodeInsert(RE.OP_STAR, 0, ret);
+                    setNextOfEnd(ret + RE.nodeSize, ret);
+                    break;
+                }
 
                 case '+':
                 {
-                    // X+ is compiled as X({gotoX}|)
-                    int branch;
-                    branch = node(RE.OP_BRANCH, 0);                   // a new branch
-                    setNextOfEnd(ret, branch);                        // is added to the end of X
-                    setNextOfEnd(node(RE.OP_GOTO, 0), ret);           // one option is to go back to the start
-                    setNextOfEnd(branch, node(RE.OP_BRANCH, 0));      // the other option
-                    setNextOfEnd(ret, node(RE.OP_NOTHING, 0));        // is OP_NOTHING
+                    nodeInsert(RE.OP_CONTINUE, 0, ret);
+                    int n = node(RE.OP_PLUS, 0);
+                    setNextOfEnd(ret + RE.nodeSize, n);
+                    setNextOfEnd(n, ret);
+                    break;
                 }
-                break;
             }
         }
         else
         {
-            // Add end after closured subexpr
-            setNextOfEnd(ret, node(RE.OP_END, 0));
-
             // Actually do the closure now
             switch (closureType)
             {
                 case '?':
+                {
                     nodeInsert(RE.OP_RELUCTANTMAYBE, 0, ret);
+                    int n = node(RE.OP_NOTHING, 0);
+                    setNextOfEnd(ret, n);
+                    setNextOfEnd(ret + RE.nodeSize, n);
                     break;
+                }
 
                 case '*':
+                {
                     nodeInsert(RE.OP_RELUCTANTSTAR, 0, ret);
+                    setNextOfEnd(ret + RE.nodeSize, ret);
                     break;
+                }
 
                 case '+':
-                    nodeInsert(RE.OP_RELUCTANTPLUS, 0, ret);
+                {
+                    nodeInsert(RE.OP_CONTINUE, 0, ret);
+                    int n = node(RE.OP_RELUCTANTPLUS, 0);
+                    setNextOfEnd(n, ret);
+                    setNextOfEnd(ret + RE.nodeSize, n);
                     break;
+                }
             }
-
-            // Point to the expr after the closure
-            setNextOfEnd(ret, lenInstruction);
         }
+
         return ret;
     }
 
     /**
-     * Compile one branch of an or operator (implements concatenation)
+     * Compile body of one branch of an or operator (implements concatenation)
+     *
      * @param flags Flags passed by reference
-     * @return Pointer to branch node
+     * @return Pointer to first node in the branch
      * @exception RESyntaxException Thrown if the regular expression has invalid syntax.
      */
     int branch(int[] flags) throws RESyntaxException
     {
         // Get each possibly closured piece and concat
         int node;
-        int ret = node(RE.OP_BRANCH, 0);
+        int ret = -1;
         int chain = -1;
         int[] closureFlags = new int[1];
         boolean nullable = true;
@@ -1213,12 +1125,15 @@ public class RECompiler
 
             // Chain starts at current
             chain = node;
+            if (ret == -1) {
+                ret = node;
+            }
         }
 
         // If we don't run loop, make a nothing node
-        if (chain == -1)
+        if (ret == -1)
         {
-            node(RE.OP_NOTHING, 0);
+            ret = node(RE.OP_NOTHING, 0);
         }
 
         // Set nullable flag for this branch
@@ -1226,12 +1141,14 @@ public class RECompiler
         {
             flags[0] |= NODE_NULLABLE;
         }
+
         return ret;
     }
 
     /**
      * Compile an expression with possible parens around it.  Paren matching
      * is done at this level so we can tie the branch tails together.
+     *
      * @param flags Flag value passed by reference
      * @return Node index of expression in instruction array
      * @exception RESyntaxException Thrown if the regular expression has invalid syntax.
@@ -1245,11 +1162,11 @@ public class RECompiler
         if ((flags[0] & NODE_TOPLEVEL) == 0 && pattern.charAt(idx) == '(')
         {
             // if its a cluster ( rather than a proper subexpression ie with backrefs )
-            if ( idx + 2 < len && pattern.charAt( idx + 1 ) == '?' && pattern.charAt( idx + 2 ) == ':' )
+            if (idx + 2 < len && pattern.charAt(idx + 1) == '?' && pattern.charAt(idx + 2) == ':')
             {
                 paren = 2;
                 idx += 3;
-                ret = node( RE.OP_OPEN_CLUSTER, 0 );
+                ret = node(RE.OP_OPEN_CLUSTER, 0);
             }
             else
             {
@@ -1260,7 +1177,8 @@ public class RECompiler
         }
         flags[0] &= ~NODE_TOPLEVEL;
 
-        // Create a branch node
+        // Process contents of first branch node
+        boolean open = false;
         int branch = branch(flags);
         if (ret == -1)
         {
@@ -1274,14 +1192,20 @@ public class RECompiler
         // Loop through branches
         while (idx < len && pattern.charAt(idx) == '|')
         {
+            // Now open the first branch since there are more than one
+            if (!open) {
+                nodeInsert(RE.OP_BRANCH, 0, branch);
+                open = true;
+            }
+
             idx++;
-            branch = branch(flags);
-            setNextOfEnd(ret, branch);
+            setNextOfEnd(branch, branch = node(RE.OP_BRANCH, 0));
+            branch(flags);
         }
 
         // Create an ending node (either a close paren or an OP_END)
         int end;
-        if ( paren > 0 )
+        if (paren > 0)
         {
             if (idx < len && pattern.charAt(idx) == ')')
             {
@@ -1291,13 +1215,13 @@ public class RECompiler
             {
                 syntaxError("Missing close paren");
             }
-            if ( paren == 1 )
+            if (paren == 1)
             {
                 end = node(RE.OP_CLOSE, closeParens);
             }
             else
             {
-                end = node( RE.OP_CLOSE_CLUSTER, 0 );
+                end = node(RE.OP_CLOSE_CLUSTER, 0);
             }
         }
         else
@@ -1310,16 +1234,16 @@ public class RECompiler
 
         // Hook the ends of each branch to the end node
         int currentNode = ret;
-        int nextNodeOffset = instruction[ currentNode + RE.offsetNext ];
+        int nextNodeOffset = instruction[currentNode + RE.offsetNext];
         // while the next node o
-        while ( nextNodeOffset != 0 && currentNode < lenInstruction )
+        while (nextNodeOffset != 0 && currentNode < lenInstruction)
         {
             // If branch, make the end of the branch's operand chain point to the end node.
-            if ( instruction[ currentNode + RE.offsetOpcode ] == RE.OP_BRANCH )
+            if (instruction[currentNode /* + RE.offsetOpcode */] == RE.OP_BRANCH)
             {
-                setNextOfEnd( currentNode + RE.nodeSize, end );
+                setNextOfEnd(currentNode + RE.nodeSize, end);
             }
-            nextNodeOffset = instruction[ currentNode + RE.offsetNext ];
+            nextNodeOffset = instruction[currentNode + RE.offsetNext];
             currentNode += nextNodeOffset;
         }
 
@@ -1345,7 +1269,6 @@ public class RECompiler
         idx = 0;                                        // Set parsing index to the first character
         lenInstruction = 0;                             // Set emitted instruction count to zero
         parens = 1;                                     // Set paren level to 1 (the implicit outer parens)
-        brackets = 0;                                   // No bracketed closures yet
 
         // Initialize pass by reference flags value
         int[] flags = { NODE_TOPLEVEL };
@@ -1432,8 +1355,8 @@ public class RECompiler
                 // Min is in the range, but max is outside
                 else if (min >= minRange[i] && min <= maxRange[i])
                 {
-                    delete(i);
                     min = minRange[i];
+                    delete(i);
                     merge(min, max);
                     return;
                 }
@@ -1441,8 +1364,8 @@ public class RECompiler
                 // Max is in the range, but min is outside
                 else if (max >= minRange[i] && max <= maxRange[i])
                 {
-                    delete(i);
                     max = maxRange[i];
+                    delete(i);
                     merge(min, max);
                     return;
                 }
@@ -1478,7 +1401,6 @@ public class RECompiler
                 if (minRange[i] >= min && maxRange[i] <= max)
                 {
                     delete(i);
-                    i--;
                     return;
                 }
 
